package org.fuseri.moduleexercise.common;

import jakarta.persistence.EntityNotFoundException;
import lombok.Getter;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;

import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;

/**
 * Dummy implementation of repository. Later will be replaced by JpaRepository.
 *
 * @param <T>  entity
 */
public abstract class DomainRepositoryImpl<T extends DomainObject> implements DomainRepository<T, String> {

    /**
     * Dummy database
     */
    @Getter
    private final Set<T> items = new HashSet<>();

    /**
     * Save the specified entity
     *
     * @param entity entity to be saved
     * @return created entity
     */
    @Override
    public T save(T entity) {
        items.add(entity);
        return entity;
    }

    /**
     * Find entity by ID
     *
     * @param id ID of entity to be found
     * @return {@code Optional} containing the found entity,
     * or an empty {@code Optional} if no such entity exists
     */
    @Override
    public Optional<T> findById(String id) {
        return items.stream()
                .filter(e -> e.getId().equals(id))
                .findFirst();
    }

    /**
     * Retrieve a page of entities according to the specified pagination information
     *
     * @param pageRequest the pagination information for the query
     * @return a page of entities that satisfy the pagination criteria
     */
    @Override
    public Page<T> findAll(PageRequest pageRequest) {

        int startIndex = pageRequest.getPageNumber() * pageRequest.getPageSize();

        List<T> pageEntities = items.stream()
                .skip(startIndex)
                .limit(pageRequest.getPageSize())
                .toList();

        return new PageImpl<>(pageEntities, pageRequest, pageEntities.size());
    }

    /**
     * Update entity
     *
     * @param entity entity to update
     * @return updated entity
     */
    @Override
    public T update(T entity) {
        if (entity == null || entity.getId() == null) {
            throw new IllegalArgumentException("Entity and its ID can not be null.");
        }

        var optionalEntity = findById(entity.getId());
        if (optionalEntity.isEmpty()) {
            throw new EntityNotFoundException("Entity not found with ID: " + entity.getId());
        }

        T oldEntity = optionalEntity.get();
        items.remove(oldEntity);
        items.add(entity);
        return entity;
    }

    /**
     * Delete the entity with the specified id
     * Note that this does not do cascade deleting.
     * We will have cascade deleting with usage of JpaRepository
     *
     * @param id the id of the entity to be deleted
     */
    @Override
    public void deleteById(String id) {
        items.removeIf(e -> e.getId().equals(id));
    }
}